/**
 * <copyright>
 * 
 * Copyright (c) 2013 itemis and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *     itemis - Initial API and implementation
 * 
 * </copyright>
 */
package org.safe.ide.ui.preferences;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ProjectScope;
import org.eclipse.core.runtime.SafeRunner;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.preference.ComboFieldEditor;
import org.eclipse.jface.preference.FieldEditor;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.util.SafeRunnable;
import org.eclipse.sphinx.emf.metamodel.IMetaModelDescriptor;
import org.eclipse.sphinx.emf.metamodel.MetaModelDescriptorRegistry;
import org.eclipse.sphinx.emf.resource.IModelConverter;
import org.eclipse.sphinx.emf.resource.IModelConverterDescription;
import org.eclipse.sphinx.emf.resource.ModelConverterRegistry;
import org.eclipse.sphinx.emf.workspace.loading.ModelLoadManager;
import org.eclipse.sphinx.emf.workspace.saving.ModelSaveManager;
import org.eclipse.sphinx.emf.workspace.ui.saving.BasicModelSaveableFilter;
import org.eclipse.sphinx.platform.ui.preferences.AbstractPreferenceAndPropertyPage;
import org.eclipse.sphinx.platform.util.ExtendedPlatform;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.preferences.ScopedPreferenceStore;
import org.safe.common.metamodel.SAFEMetaModelVersionData;
import org.safe.common.metamodel.SAFEReleaseDescriptor;
import org.safe.ide.natures.SAFENature;
import org.safe.ide.preferences.ISAFEWorkspacePreferenceConstants;
import org.safe.ide.ui.internal.Activator;
import org.safe.ide.ui.internal.messages.Messages;
import org.safe.ide.ui.internal.preferences.ModelConverterDescriptionLabels;
import org.safe.ide.ui.messages.WorkspaceUIMessages;

public class ReleasePreferencePage extends AbstractPreferenceAndPropertyPage {

	private String LABEL_PATTERN = "%1s (release %2s)"; //$NON-NLS-1$

	private static final String PROP_PAGE_ID = "org.safe.ide.ui.propertyPages.safe.release"; //$NON-NLS-1$

	private static final String VERSION_SEPARATOR = "."; //$NON-NLS-1$

	private static final String OTHER_RESOURCE_VERSION_VALUE = "OTHER"; //$NON-NLS-1$

	private static final String DEFAULT_RESOURCE_VERSION = null;

	// Release fields
	private String initialRelease;
	private Group releaseGroup;
	private ComboFieldEditor releaseField;
	private boolean releaseValueChanged = false;

	// Resource Version fields
	private String initialResourceVersion;
	private boolean isResourceVersionSameAsOriginal;
	private Composite resourceVersionRadioBox;
	private Label resourceVersionLabel;
	private Combo resourceVersionCombo;
	private ModelConverterDescriptionLabels converterDescriptionLabels;

	final Map<String, SAFEReleaseDescriptor> resourceVersionNamesAndValues = new HashMap<String, SAFEReleaseDescriptor>();

	public ReleasePreferencePage() {
		super(SAFENature.ID, GRID);
	}

	public ReleasePreferencePage(String requiredProjectNatureId, int style) {
		super(requiredProjectNatureId, style);
	}

	@Override
	protected String getPreferencePageID() {
		return ISAFEWorkspacePreferenceConstants.SAFE_RELEASE_PREFERENCE_PAGE_ID;
	}

	@Override
	protected String getPropertyPageID() {
		return PROP_PAGE_ID;
	}

	@Override
	protected IPreferenceStore doGetPreferenceStore() {
		if (isProjectPreferencePage()) {
			return getProjectPreferenceStore((IProject) getElement());
		} else {
			return getWorkspacePreferenceStore();
		}
	}

	@Override
	protected void addFields(Composite parent) {
		addReleaseGroup(parent);
	}

	@Override
	protected void initialize() {
		super.initialize();

		updateRelease(initialRelease);

		if (isProjectPreferencePage()) {
			updateResourceVersionWidgets();
			updateResourceVersionCombo(initialRelease, initialResourceVersion);
		}
	}

	@Override
	protected void enablePreferenceContent(boolean useProjectSpecificSettings) {
		if (releaseField != null) {
			releaseField.setEnabled(useProjectSpecificSettings, releaseGroup);
		}
		if (resourceVersionLabel != null) {
			resourceVersionLabel.setEnabled(useProjectSpecificSettings);
		}
		if (resourceVersionRadioBox != null) {
			setEnabledDeep(resourceVersionRadioBox, useProjectSpecificSettings);
			updateResourceVersionWidgets();
		}
	}

	@Override
	public void propertyChange(PropertyChangeEvent event) {
		super.propertyChange(event);

		if (isProjectPreferencePage()) {
			if (event.getSource().equals(releaseField) && event.getProperty().equals(FieldEditor.VALUE)) {
				releaseValueChanged = event.getNewValue().equals(initialRelease) == false;
				updateResourceVersionCombo((String) event.getNewValue(), DEFAULT_RESOURCE_VERSION);
			}
		}
	}

	@Override
	public boolean performOk() {
		super.performOk();

		if (!isProjectPreferencePage()) {
			return true;
		}

		if (useProjectSettings()) {
			String selectedResourceVersion;

			if (isResourceVersionSameAsOriginal) {
				selectedResourceVersion = ISAFEWorkspacePreferenceConstants.PREF_VALUE_AUTO_CONVERSION_MODE;
			} else {
				selectedResourceVersion = getSelectedResourceVersion().getIdentifier();
			}

			getPreferenceStore().setValue(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION, selectedResourceVersion);
		} else {
			getPreferenceStore().setToDefault(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE);
			getPreferenceStore().setToDefault(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION);

			// determine wether a reload is necessary based on workspace level preference
			IPreferenceStore workspacePreferenceStore = getWorkspacePreferenceStore();

			releaseValueChanged = initialRelease.equals(workspacePreferenceStore.getString(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE)) == false;
		}

		// Release is about to change?
		if (releaseValueChanged) {
			// Give user a chance to save pending changes on SAFE model with old release
			final boolean canceled[] = new boolean[1];
			final IProject project = (IProject) getElement();
			PlatformUI.getWorkbench().getDisplay().syncExec(new Runnable() {
				public void run() {
					SafeRunner.run(new SafeRunnable(Messages.error_failedToSaveModelsInWorkbench) {
						public void run() {
							IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
							if (window == null) {
								IWorkbenchWindow[] windows = PlatformUI.getWorkbench().getWorkbenchWindows();
								if (windows.length > 0) {
									window = windows[0];
								}
							}
							if (window != null) {
								canceled[0] = !PlatformUI.getWorkbench().saveAll(window, window, new BasicModelSaveableFilter(project), true);
							} else {
								// TODO Save only SAFE model
								ModelSaveManager.INSTANCE.saveProject(project, false, null);
								canceled[0] = false;
							}
						}
					});
				}
			});

			// Abort preference change if saving has been canceled by user
			if (canceled[0]) {
				return false;
			}

			// Force reset of dirty information on all models in given project for clearing dirty information of those
			// models that have not been taken into account by the save operation (happens e.g. when user deselects some
			// or all of them before proceeding with the save operation)
			Collection<IProject> projectGroup = ExtendedPlatform.getProjectGroup(project, true);
			for (IProject proj : projectGroup) {
				ModelSaveManager.INSTANCE.setSaved(proj);
			}

			// Schedule unload of SAFE model with old release and load of SAFE model with new release as a
			// background operation
			ModelLoadManager.INSTANCE.reloadProject(project, false, SAFEReleaseDescriptor.INSTANCE, true, null);
		}

		return true;
	}

	private void addReleaseGroup(final Composite parent) {
		getWorkspacePreferenceStore().setDefault(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE,
				ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE_DEFAULT);

		if (isProjectPreferencePage()) {
			// treat the missing project release preference based on what is in effect on the workspace level
			getPreferenceStore().setDefault(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE,
					getWorkspacePreferenceStore().getString(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE));
		}

		initialRelease = getPreferenceStore().getString(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE);

		releaseGroup = new Group(parent, SWT.NONE);
		releaseGroup.setText(WorkspaceUIMessages.ReleasePreferencePage_releaseGroupText);

		// The list of supported SAFE meta-model releases
		List<SAFEReleaseDescriptor> supportedReleases = MetaModelDescriptorRegistry.INSTANCE.getDescriptors(SAFEReleaseDescriptor.INSTANCE, true);

		String[][] entryNamesAndValues = new String[supportedReleases.size()][2];
		for (int index = 0; index < supportedReleases.size(); index++) {
			SAFEReleaseDescriptor safeRelease = supportedReleases.get(index);
			entryNamesAndValues[index][0] = String.format(LABEL_PATTERN, safeRelease.getName(), getVersionNumber(safeRelease));
			entryNamesAndValues[index][1] = safeRelease.getIdentifier();
		}

		releaseField = new ComboFieldEditor(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE,
				WorkspaceUIMessages.ReleasePreferencePage_releaseFieldLabel, entryNamesAndValues, releaseGroup);
		addField(releaseField);

		// set layout of release group after field addition since it is being altered when field editors are added
		if (isProjectPreferencePage()) {
			GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
			gridData.horizontalSpan = 2;
			releaseGroup.setLayoutData(gridData);

			GridLayout gridLayout = new GridLayout(2, false);
			releaseGroup.setLayout(gridLayout);
		} else {
			GridData gridData = new GridData(SWT.FILL, SWT.TOP, true, false);
			gridData.horizontalSpan = 2;
			releaseGroup.setLayoutData(gridData);

			GridLayout gridLayout = new GridLayout(2, false);
			releaseGroup.setLayout(gridLayout);
		}

		if (isProjectPreferencePage()) {
			if (getPreferenceStore().isDefault(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION)) {
				initialResourceVersion = ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION_DEFAULT;
			} else {
				initialResourceVersion = getPreferenceStore().getString(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION);
			}
			isResourceVersionSameAsOriginal = initialResourceVersion.equals(ISAFEWorkspacePreferenceConstants.PREF_VALUE_AUTO_CONVERSION_MODE);

			resourceVersionLabel = new Label(releaseGroup, SWT.LEFT);
			resourceVersionLabel.setText(WorkspaceUIMessages.ReleasePreferencePage_resourceVersionLabel);
			{
				GridData gridData = new GridData(SWT.FILL, SWT.TOP, true, false);
				gridData.horizontalSpan = 2;
				resourceVersionLabel.setLayoutData(gridData);
			}

			resourceVersionRadioBox = new Composite(releaseGroup, SWT.NONE);
			{
				GridData gridData = new GridData(SWT.FILL, SWT.TOP, true, false);
				gridData.horizontalSpan = 2;
				resourceVersionRadioBox.setLayoutData(gridData);

				GridLayout gridLayout = new GridLayout();
				gridLayout.marginWidth = 0;
				gridLayout.marginHeight = 0;
				gridLayout.horizontalSpacing = 8;
				resourceVersionRadioBox.setLayout(gridLayout);
			}

			createResourceVersionRadioButton(WorkspaceUIMessages.ReleasePreferencePage_sameAsInOriginalButtonLabel,
					ISAFEWorkspacePreferenceConstants.PREF_SAFE_RESOURCE_VERSION_DEFAULT, isResourceVersionSameAsOriginal);
			createResourceVersionRadioButton(WorkspaceUIMessages.ReleasePreferencePage_otherResourceVersionButtonLabel, OTHER_RESOURCE_VERSION_VALUE,
					isResourceVersionSameAsOriginal == false);

			resourceVersionCombo = new Combo(releaseGroup, SWT.READ_ONLY);
			{
				GridData gridData = new GridData(SWT.FILL, SWT.TOP, true, false);
				gridData.horizontalSpan = 2;
				gridData.horizontalIndent = convertHorizontalDLUsToPixels(16);
				resourceVersionCombo.setLayoutData(gridData);
			}
			resourceVersionCombo.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					updateConverterDescriptionArea();
				}
			});

			ScrolledComposite scrolledComposite = new ScrolledComposite(releaseGroup, SWT.BORDER | SWT.V_SCROLL);
			{
				GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
				gridData.horizontalSpan = 2;
				gridData.horizontalIndent = convertHorizontalDLUsToPixels(16);
				scrolledComposite.setLayoutData(gridData);
			}

			converterDescriptionLabels = new ModelConverterDescriptionLabels(scrolledComposite);
			scrolledComposite.setContent(converterDescriptionLabels);
			scrolledComposite.setMinHeight(convertHeightInCharsToPixels(3));
			scrolledComposite.setAlwaysShowScrollBars(true);
		}

		Dialog.applyDialogFont(releaseGroup);
	}

	private String getVersionNumber(SAFEReleaseDescriptor releaseDesc) {
		SAFEMetaModelVersionData versionData = releaseDesc.getSAFEVersionData();
		StringBuilder versionNumber = new StringBuilder();
		versionNumber.append(versionData.getMajor());
		versionNumber.append(VERSION_SEPARATOR);
		versionNumber.append(versionData.getMinor());
		versionNumber.append(VERSION_SEPARATOR);
		versionNumber.append(versionData.getRevision());
		return versionNumber.toString();
	}

	private void updateRelease(String safeReleaseId) {
		// hack to trick combo field into selecting the different value
		// TODO replace combo field with custom combo
		String value = getPreferenceStore().getString(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE);
		getPreferenceStore().setValue(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE, safeReleaseId);
		releaseField.load();
		getPreferenceStore().setValue(ISAFEWorkspacePreferenceConstants.PREF_SAFE_RELEASE, value);
	}

	private SAFEReleaseDescriptor getSelectedResourceVersion() {
		int selectedIndex = resourceVersionCombo.getSelectionIndex();

		if (selectedIndex == -1) {
			return null;
		}

		String selectedName = resourceVersionCombo.getItem(selectedIndex);
		return resourceVersionNamesAndValues.get(selectedName);
	}

	private void updateResourceVersionCombo(IMetaModelDescriptor safeRelease, String resourceVersion) {
		List<SAFEReleaseDescriptor> relevantVersions = new ArrayList<SAFEReleaseDescriptor>();

		List<IModelConverter> converters = ModelConverterRegistry.INSTANCE.getConverters();
		for (IModelConverter modelConverter : converters) {
			// Check if meta-model version matches
			if (modelConverter.getMetaModelVersionDescriptor().equals(safeRelease)) {
				relevantVersions.add((SAFEReleaseDescriptor) modelConverter.getResourceVersionDescriptor());
			}
		}

		Collections.sort(relevantVersions);

		String[] resourceVersionNames = new String[relevantVersions.size()];
		int i = 0;
		int selectedIndex = 0;
		for (SAFEReleaseDescriptor relevantVersion : relevantVersions) {
			String name = "SAFE " + getVersionNumber(relevantVersion); //$NON-NLS-1$
			resourceVersionNames[i] = name;
			String identifier = relevantVersion.getIdentifier();
			resourceVersionNamesAndValues.put(name, relevantVersion);
			// Used to select the serialized identifier in the Combo
			if (identifier.equals(resourceVersion)) {
				selectedIndex = i;
			}
			i++;
		}
		resourceVersionCombo.setItems(resourceVersionNames);

		if (relevantVersions.size() > 0) {
			resourceVersionCombo.select(selectedIndex);
		}

		updateConverterDescriptionArea();
	}

	private void updateResourceVersionCombo(String safeReleaseId, String resourceVersion) {
		IMetaModelDescriptor safeRelease = MetaModelDescriptorRegistry.INSTANCE.getDescriptor(safeReleaseId);
		updateResourceVersionCombo(safeRelease, resourceVersion);
	}

	private void updateConverterDescriptionArea() {
		if (isResourceVersionSameAsOriginal == false) {
			SAFEReleaseDescriptor safeRelease = getSelectedResourceVersion();

			if (safeRelease != null) {
				Collection<IModelConverterDescription> convDescs = ModelConverterRegistry.INSTANCE
						.getConverterDescriptions(getSelectedResourceVersion());
				if (convDescs.size() > 0) {
					converterDescriptionLabels.setConverterDescription(convDescs.iterator().next());
					setEnabledDeep(converterDescriptionLabels, true);
					return;
				}
			}
		}

		converterDescriptionLabels.setConverterDescription(null);
		setEnabledDeep(converterDescriptionLabels, false);
	}

	private void updateResourceVersionWidgets() {
		boolean enabled = isResourceVersionSameAsOriginal == false && useProjectSettings();
		resourceVersionCombo.setEnabled(enabled);
		setEnabledDeep(converterDescriptionLabels, enabled);
	}

	private Button createResourceVersionRadioButton(String label, String value, boolean selected) {
		Button radio = new Button(resourceVersionRadioBox, SWT.RADIO | SWT.LEFT);
		radio.setText(label);
		radio.setData(value);
		radio.setSelection(selected);
		radio.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				String value = (String) event.widget.getData();
				isResourceVersionSameAsOriginal = value.equals(ISAFEWorkspacePreferenceConstants.PREF_VALUE_AUTO_CONVERSION_MODE);
				updateResourceVersionWidgets();
				updateConverterDescriptionArea();
			}
		});
		return radio;
	}

	private void setEnabledDeep(Composite parent, boolean isEnabled) {
		parent.setEnabled(isEnabled);
		for (Control c : parent.getChildren()) {
			c.setEnabled(isEnabled);
		}
	}

	private IPreferenceStore getProjectPreferenceStore(IProject project) {
		return new ScopedPreferenceStore(new ProjectScope(project), org.safe.ide.internal.Activator.getPlugin().getBundle().getSymbolicName());
	}

	private IPreferenceStore getWorkspacePreferenceStore() {
		return Activator.getPlugin().getWorkspacePreferenceStore();
	}
}
