/**
 * <copyright>
 * 
 * Copyright (c) 2013 itemis and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *     itemis - Initial API and implementation
 * 
 * </copyright>
 */
package org.safe.serialization.internal;

import static org.eclipse.sphinx.emf.resource.ExtendedResource.OPTION_SCHEMA_LOCATION_CATALOG;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eatop.eel.serialization.internal.ExtendedXMLPersistenceMappingSaveImpl;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.emf.ecore.util.InternalEList;
import org.eclipse.emf.ecore.xmi.XMLHelper;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xmi.impl.XMLString;
import org.eclipse.emf.ecore.xml.namespace.XMLNamespacePackage;
import org.eclipse.sphinx.emf.Activator;
import org.eclipse.sphinx.emf.metamodel.IMetaModelDescriptor;
import org.eclipse.sphinx.emf.metamodel.MetaModelDescriptorRegistry;
import org.eclipse.sphinx.emf.resource.ExtendedResource;
import org.eclipse.sphinx.emf.resource.IModelConverter;
import org.eclipse.sphinx.platform.util.PlatformLogUtil;
import org.safe.common.metamodel.SAFEMetaModelVersionData;
import org.safe.common.metamodel.SAFEReleaseDescriptor;
import org.safe.common.util.ISAFEConstants;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * An XMLSaveImpl that is specially designed to load SAFE resources.
 */
public class SafeXMLPersistenceMappingSaveImpl extends ExtendedXMLPersistenceMappingSaveImpl {

	private static final String SCHEMA_BASE = "safe"; //$NON-NLS-1$
	private static final String SCHEMA_EXTENSION = "xsd"; //$NON-NLS-1$

	/**
	 * Creates a new XMLSaveImpl associated with given helper.
	 * 
	 * @param helper
	 *            The helper that provides info about the way serialization should be performed.
	 */
	public SafeXMLPersistenceMappingSaveImpl(XMLHelper helper) {
		super(helper);
	}

	// FIXME: to check 2013-01-09
	@Override
	/**
	 * Saves the type attribute. 
	 * If the package of eClass is other than the package of the feature's containing class, 
	 * then the name of the eClass is added to be saved. 
	 * To avoid the validation problem, the prefix is not saved directly in the type attribute, 
	 * otherwise the "prefix:name", e.g., "_1:HW_ELEMENT" could not be validated by the parser.
	 * The mapping of prefixToURI should be updated internally.
	 */
	protected void saveTypeAttribute(EStructuralFeature feature, EClass eClass) {
		String typeAttribute = xmlPersistenceMappingExtendedMetaData.getXMLTypeAttributeName(feature);
		if (null == typeAttribute || "".equals(typeAttribute) || XSI_TYPE_NS.equals(typeAttribute)) { //$NON-NLS-1$
			declareXSI = true;
			typeAttribute = XSI_TYPE_NS;
			doc.addAttribute(typeAttribute, helper.getQName(eClass));
		} else {
			if (feature.getEContainingClass().getEPackage() != eClass.getEPackage()) {
				// To avoid saving the prefix in the type attribute, save the class name instead, and update the
				// prefixToURI mapping internal
				doc.addAttribute(typeAttribute, helper.getName(eClass));
				((SafeXMLPersistenceMappingHelperImpl) helper).updatePrefixToURI(eClass.getEPackage(), false);
			} else {
				doc.addAttribute(typeAttribute, helper.getName(eClass));
			}
		}
	}

	protected void saveReferencedHREFWithTypeSaved(EStructuralFeature f, EObject remote, String qname, boolean doSaveType) {
		{
			String href = helper.getHREF(remote);
			if (href != null) {
				href = convertURI(href);
				EClass eClass = remote.eClass();
				doc.startElement(qname);
				saveTypeAttribute(f, eClass);
				doc.endContentElement(href);
			}
		}
	}

	// TODO TO BE CHECKED
	/**
	 * Save EReference Referenced 1100 many. The type attribute is always saved instead of saved depends on the
	 * conditions which is the default case.
	 * 
	 * @param values
	 * @param f
	 */
	@Override
	protected void saveEReferenceReferenced1100Many(InternalEList<? extends EObject> values, EStructuralFeature f) {
		int size = values.size();
		String qname;
		EObject value;
		doc.startElement(getFeatureWrapperQName(f));
		for (int i = 0; i < size; i++) {
			value = values.basicGet(i);
			qname = getFeatureQName(f);
			saveReferencedHREFWithTypeSaved(f, value, qname, true);
		}
		doc.endElement();
	}

	// TODO TO BE CHECKED
	@Override
	/**
	 * Save EReference Referenced 0100 single. The type attribute is always saved instead of saving depending on the
	 * conditions (remote.eClass() != (EClass) f.getEType()) which is the default case.
	 *
	 * @param values
	 * @param f
	 */
	protected void saveEReferenceReferenced0100Single(EObject remote, EStructuralFeature f) {
		String qname = getFeatureQName(f);
		saveReferencedHREFWithTypeSaved(f, remote, qname, true);
	}

	/*
	 * @see org.eclipse.emf.ecore.xmi.impl.XMLSaveImpl#addNamespaceDeclarations()
	 */
	@Override
	protected void addNamespaceDeclarations() {
		// This is a Hack! Sphinx ExtendedXMLSaveImpl assumes that different revisions have different namespaces. This
		// is not true for SAFE 1.0. Therefore the Converter, SchemaLocationCatalog and SchemaLocationMap are
		// manipulated in such a way that the ExtendedXMLSaveImpl will do what we expect even for 1.0. This can be
		// removed as soon as Sphinx mechanism is refactored to be MetaModelDescriptor based instead of namespace based.
		IMetaModelDescriptor release = MetaModelDescriptorRegistry.INSTANCE.getDescriptor(helper.getResource());
		boolean isSingleNsReleaseResource = isSingleNsRelease((SAFEReleaseDescriptor) release);
		@SuppressWarnings("unchecked")
		Map<String, String> originalSchemaCatalog = (Map<String, String>) getDefaultSaveOptions().get(OPTION_SCHEMA_LOCATION_CATALOG);
		IModelConverter originalConverter = converter;
		if (isSingleNsReleaseResource) {
			if (converter != null) {
				if (converter.getResourceVersionDescriptor() != release) {
					setSchemaLocationFromConverter();
				}
			} else {
				Map<String, String> schemaLocationMap = getSchemaLocationMap();
				if (!schemaLocationMap.isEmpty()) {
					setSchemaCatalog(schemaLocationMap);
					converter = new SingleNsReleaseConverter(release);
				}
			}
		}

		// ========= super.addNamespaceDeclarations();
		EPackage noNamespacePackage = helper.getNoNamespacePackage();
		EPackage[] packages = helper.packages();
		StringBuffer buffer = getBuffer();
		if (buffer == null) {
			buffer = new StringBuffer();
		}
		buffer.setLength(0);
		StringBuffer xsiSchemaLocation = buffer;
		String xsiNoNamespaceSchemaLocation = null;
		if (declareSchemaLocation) {
			Map<String, String> handledBySchemaLocationMap = new HashMap<String, String>();
			@SuppressWarnings("unchecked")
			Map<String, String> schemaLocationCatalog = (Map<String, String>) options.get(ExtendedResource.OPTION_SCHEMA_LOCATION_CATALOG);

			if (extendedMetaData != null) {
				Resource resource = helper.getResource();
				if (resource != null && resource.getContents().size() >= 1) {
					EObject schemaLocationRoot = getSchemaLocationRoot(resource.getContents().get(0));

					EReference xsiSchemaLocationMapFeature = extendedMetaData.getXSISchemaLocationMapFeature(schemaLocationRoot.eClass());
					if (xsiSchemaLocationMapFeature != null) {
						@SuppressWarnings("unchecked")
						EMap<String, String> xsiSchemaLocationMap = (EMap<String, String>) schemaLocationRoot.eGet(xsiSchemaLocationMapFeature);
						if (!xsiSchemaLocationMap.isEmpty()) {
							// Write schema location value string from recorded schema location entries
							handledBySchemaLocationMap = new HashMap<String, String>(xsiSchemaLocationMap.map());
							for (Map.Entry<String, String> entry : xsiSchemaLocationMap.entrySet()) {
								String namespace = entry.getKey();
								String location = entry.getValue();
								URI locationURI = URI.createURI(location);
								if (namespace == null) {
									declareXSI = true;

									xsiNoNamespaceSchemaLocation = helper.deresolve(locationURI).toString();
								} else {
									// Need to adjust current schema namespace according to applicable model converter?
									if (converter != null) {
										String mmBaseNsURI = converter.getMetaModelVersionDescriptor().getNamespace();
										String resourceBaseNsURI = converter.getResourceVersionDescriptor().getNamespace();
										if (mmBaseNsURI != null && resourceBaseNsURI != null && namespace.startsWith(mmBaseNsURI)) {
											// Substitute metamodel version in current schema namespace with
											// resource version of applicable model converter
											namespace = namespace.replace(mmBaseNsURI, resourceBaseNsURI);

											// Retrieve schema location corresponding to adjusted schema namespace
											// from schema location catalog
											locationURI = null;
											if (schemaLocationCatalog != null) {
												location = schemaLocationCatalog.get(namespace);
												if (location != null) {
													locationURI = URI.createURI(location);
												} else {
													PlatformLogUtil.logAsWarning(Activator.getPlugin(), new RuntimeException(
															"Schema location catalog entry for namespace '" + namespace + "' is missing")); //$NON-NLS-1$ //$NON-NLS-2$
												}
											} else {
												PlatformLogUtil.logAsWarning(Activator.getPlugin(), new RuntimeException(
														"Schema location catalog is missing")); //$NON-NLS-1$
											}
										}
									} else {
										// Current schema namespace corresponding to a metamodel version that is an
										// older version of the metamodel behind schema location root?
										IMetaModelDescriptor mmDescriptor = MetaModelDescriptorRegistry.INSTANCE.getDescriptor(schemaLocationRoot);
										if (mmDescriptor != null) {
											boolean compatible = false;
											for (java.net.URI compatibleNsURI : mmDescriptor.getCompatibleNamespaceURIs()) {
												if (namespace.startsWith(compatibleNsURI.toString())) {
													compatible = true;
													break;
												}
											}
											if (compatible) {
												// Ignore corresponding recorded schema location entry; as no model
												// converter is available the model will be serialized using the most
												// recent metamodel namespace; we therefore must not write a schema
												// location entry for the original old namespace but the most recent
												// one; this is done below anyway, so we can simply skip the rest right
												// here
												continue;
											}
										}
									}

									if (locationURI != null) {
										declareXSI = true;

										if (xsiSchemaLocation.length() > 0) {
											xsiSchemaLocation.append(' ');
										}
										xsiSchemaLocation.append(namespace);
										xsiSchemaLocation.append(' ');
										xsiSchemaLocation.append(helper.deresolve(locationURI).toString());
									}
								}
							}
						}
					}
				}
			}

			// Complete schema location value string with calculated schema location entries
			for (EPackage ePackage : packages) {

				String javaImplementationLocation = null;
				if (declareSchemaLocationImplementation) {
					// First try to see if this package's implementation class has an eInstance.
					//
					try {
						Field field = ePackage.getClass().getField("eINSTANCE"); //$NON-NLS-1$
						javaImplementationLocation = "java://" + field.getDeclaringClass().getName(); //$NON-NLS-1$
					} catch (Exception exception) {
						// If there is no field, then we can't do this.
					}
				}

				if (noNamespacePackage == ePackage) {
					if (ePackage.eResource() != null && !handledBySchemaLocationMap.containsKey(null)) {
						if (javaImplementationLocation != null) {
							xsiNoNamespaceSchemaLocation = javaImplementationLocation;
						} else if (schemaLocationCatalog != null) {
							// Retrieve schema location corresponding to no namespace from schema location catalog
							xsiNoNamespaceSchemaLocation = schemaLocationCatalog.get(null);
							if (xsiNoNamespaceSchemaLocation == null) {
								PlatformLogUtil.logAsWarning(Activator.getPlugin(), new RuntimeException(
										"Schema location catalog entry for no namespace (null) is missing")); //$NON-NLS-1$ 
							}
						} else {
							xsiNoNamespaceSchemaLocation = helper.getHREF(ePackage);
							if (xsiNoNamespaceSchemaLocation != null && xsiNoNamespaceSchemaLocation.endsWith("#/")) { //$NON-NLS-1$
								xsiNoNamespaceSchemaLocation = xsiNoNamespaceSchemaLocation.substring(0, xsiNoNamespaceSchemaLocation.length() - 2);
							}
						}

						declareXSI = xsiNoNamespaceSchemaLocation != null;
					}
				} else
				// FIXME Retrieve schema location only for the SAFE packages
				if (ePackage.getNsURI().contains(ISAFEConstants.SAFE_PACKAGE_NSURI)) {
					String namespace = extendedMetaData == null ? ePackage.getNsURI() : extendedMetaData.getNamespace(ePackage);

					// Need to adjust schema namespace according to applicable model converter?
					if (converter != null) {
						String mmBaseNsURI = converter.getMetaModelVersionDescriptor().getNamespace();
						String resourceBaseNsURI = converter.getResourceVersionDescriptor().getNamespace();
						if (mmBaseNsURI != null && resourceBaseNsURI != null && namespace.startsWith(mmBaseNsURI)) {
							// Substitute metamodel version in schema namespace with resource
							// version of applicable model converter
							namespace = namespace.replace(mmBaseNsURI, resourceBaseNsURI);
						}
					}

					if (!handledBySchemaLocationMap.containsKey(namespace)) {
						if (javaImplementationLocation != null || namespace != null) {
							String location = null;
							if (javaImplementationLocation != null) {
								location = javaImplementationLocation;
							} else if (schemaLocationCatalog != null) {
								// Retrieve schema location corresponding to schema namespace from schema location
								// catalog
								location = schemaLocationCatalog.get(namespace);
								if (location == null) {
									PlatformLogUtil.logAsWarning(Activator.getPlugin(), new RuntimeException(
											"Schema location catalog entry for namespace '" + namespace + "' is missing")); //$NON-NLS-1$ //$NON-NLS-2$
								}
							} else {
								location = helper.getHREF(ePackage);
								location = convertURI(location);
								if (location.endsWith("#/")) { //$NON-NLS-1$
									location = location.substring(0, location.length() - 2);
									URI uri = ePackage.eResource().getURI();
									if (uri != null && uri.hasFragment()) {
										location += "#" + uri.fragment(); //$NON-NLS-1$
									}
								}
							}

							if (location != null) {
								declareXSI = true;

								if (xsiSchemaLocation.length() > 0) {
									xsiSchemaLocation.append(' ');
								}
								xsiSchemaLocation.append(namespace);
								xsiSchemaLocation.append(' ');
								xsiSchemaLocation.append(location);
							}

							// Avoid duplicate schema location entries
							handledBySchemaLocationMap.put(namespace, location);
						}
					}
				}
			}
		}

		for (EPackage ePackage : packages) {
			if (ePackage != noNamespacePackage && ePackage != XMLNamespacePackage.eINSTANCE
					&& !ExtendedMetaData.XMLNS_URI.equals(ePackage.getNsURI())) {
				// FIXME For AUTOSAR: get AUTOSAR namespace instead of the package namespace
				// String nsURI = extendedMetaData == null ? ePackage.getNsURI() :
				// extendedMetaData.getNamespace(ePackage);
				String nsURI = getRootPackageNsURI(ePackage);

				if (ePackage == xmlSchemaTypePackage) {
					nsURI = XMLResource.XML_SCHEMA_URI;
				}
				if (nsURI != null && !isDuplicateURI(nsURI)) {
					// Need to adjust EPackage namespace according to applicable model converter?
					if (converter != null) {
						String mmBaseNsURI = converter.getMetaModelVersionDescriptor().getNamespace();
						String resourceBaseNsURI = converter.getResourceVersionDescriptor().getNamespace();
						if (mmBaseNsURI != null && resourceBaseNsURI != null && nsURI.startsWith(mmBaseNsURI)) {
							// Substitute metamodel version in EPackage namespace with resource version of applicable
							// model converter
							nsURI = nsURI.replace(mmBaseNsURI, resourceBaseNsURI);
						}
					}

					List<String> nsPrefixes = helper.getPrefixes(ePackage);
					for (String nsPrefix : nsPrefixes) {
						if (!toDOM) {
							if (nsPrefix != null && nsPrefix.length() > 0) {
								if (!declareXSI || !"xsi".equals(nsPrefix)) { //$NON-NLS-1$
									doc.addAttributeNS(XMLResource.XML_NS, nsPrefix, nsURI);
								}
							} else {
								doc.addAttribute(XMLResource.XML_NS, nsURI);
							}
						} else {
							if (nsPrefix != null && nsPrefix.length() > 0) {
								if (!declareXSI || !"xsi".equals(nsPrefix)) { //$NON-NLS-1$
									((Element) currentNode).setAttributeNS(ExtendedMetaData.XMLNS_URI, XMLResource.XML_NS + ":" + nsPrefix, nsURI); //$NON-NLS-1$
								}
							} else {
								((Element) currentNode).setAttributeNS(ExtendedMetaData.XMLNS_URI, XMLResource.XML_NS, nsURI);
							}
						}
					}
				}
			}
		}

		if (declareXSI) {
			if (!toDOM) {
				doc.addAttribute(XSI_XMLNS, XMLResource.XSI_URI);
			} else {
				((Element) currentNode).setAttributeNS(ExtendedMetaData.XMLNS_URI, XSI_XMLNS, XMLResource.XSI_URI);
			}
		}

		if (xsiSchemaLocation.length() > 0) {
			if (!toDOM) {
				doc.addAttribute(XSI_SCHEMA_LOCATION, xsiSchemaLocation.toString());
			} else {
				((Element) currentNode).setAttributeNS(XMLResource.XSI_URI, XSI_SCHEMA_LOCATION, xsiSchemaLocation.toString());
			}
		}

		if (xsiNoNamespaceSchemaLocation != null) {
			if (!toDOM) {
				doc.addAttribute(XSI_NO_NAMESPACE_SCHEMA_LOCATION, xsiNoNamespaceSchemaLocation);
			} else {
				((Element) currentNode).setAttributeNS(XMLResource.XSI_URI, XSI_NO_NAMESPACE_SCHEMA_LOCATION, xsiNoNamespaceSchemaLocation);
			}
		}
		// ======== end of super.addNamespaceDeclarations();

		// Hack part 2. Restoring manipulated values. SchemaLocationMap is not restored as it should be adjusted as soon
		// as a Converter is applied.
		if (isSingleNsReleaseResource) {
			setSchemaCatalog(originalSchemaCatalog);
			converter = originalConverter;
		}
	}

	@Override
	protected String getRootPackageNsURI(EPackage ePackage) {
		String namespace = ePackage.getNsURI();

		if (extendedMetaData != null) {
			if (namespace.contains(ISAFEConstants.AUTOSAR_ROOT_PACKAGE_NSURI)) {
				namespace = ISAFEConstants.AUTOSAR_NSURI;
			} else {
				namespace = extendedMetaData.getNamespace(ePackage);
			}
		}
		return namespace;
	}

	// ====== The following are copied from EAST-ADL
	@SuppressWarnings("unchecked")
	private Map<String, String> getSchemaLocationMap() {
		EObject schemaLocationRoot = getSchemaLocationRoot(helper.getResource().getContents().get(0));
		EReference xsiSchemaLocationMapFeature = extendedMetaData.getXSISchemaLocationMapFeature(schemaLocationRoot.eClass());
		if (xsiSchemaLocationMapFeature != null) {
			return ((EMap<String, String>) schemaLocationRoot.eGet(xsiSchemaLocationMapFeature)).map();
		}
		return null;
	}

	private class SingleNsReleaseConverter implements IModelConverter {

		private IMetaModelDescriptor fRelease;

		public SingleNsReleaseConverter(IMetaModelDescriptor release) {
			fRelease = release;
		}

		public IMetaModelDescriptor getMetaModelVersionDescriptor() {
			return fRelease;
		}

		public IMetaModelDescriptor getResourceVersionDescriptor() {
			return fRelease;
		}

		public boolean isLoadConverterFor(XMLResource resource, Map<?, ?> options) {
			return false;
		}

		public boolean isSaveConverterFor(XMLResource resource, Map<?, ?> options) {
			return false;
		}

		public InputSource convertLoad(XMLResource resource, InputStream inputStream, Map<?, ?> options) throws IOException {
			return null;
		}

		public void convertSave(XMLString xml, int flushThreshold, org.eclipse.emf.common.util.URI uri, OutputStream outputStream, String encoding,
				XMLHelper helper, Map<?, ?> options) throws IOException {

		}

		public void addExtraAttributesToSavedRootElement(XMLString rootElement, Map<?, ?> options) {

		}

		public void dispose() {

		}
	}

	private void setSchemaLocationFromConverter() {
		SAFEReleaseDescriptor revision = (SAFEReleaseDescriptor) converter.getResourceVersionDescriptor();
		Map<String, String> converterSchemaCatalog = createSchemaCatalog(revision);
		setSchemaCatalog(converterSchemaCatalog);
		// Update SchemaMap
		Map<String, String> schemaLocationMap = getSchemaLocationMap();
		schemaLocationMap.put(revision.getNamespace(), converterSchemaCatalog.get(revision.getNamespace()));
	}

	private void setSchemaCatalog(Map<String, String> schemaCatalog) {
		getDefaultSaveOptions().put(OPTION_SCHEMA_LOCATION_CATALOG, schemaCatalog);
	}

	private boolean isSingleNsRelease(SAFEReleaseDescriptor release) {
		if (release == null) {
			return false;
		}
		List<java.net.URI> revisionUris = release.getCompatibleNamespaceURIs();
		if (revisionUris.isEmpty()) {
			return false;
		}
		for (java.net.URI revisionUri : revisionUris) {
			if (!revisionUri.equals(release.getNamespaceURI())) {
				return false;
			}
		}
		return true;
	}

	private Map<Object, Object> getDefaultSaveOptions() {
		return helper.getResource().getDefaultSaveOptions();
	}

	// TODO: These constants and methods are a HACK!!!! SAFE 1.0 revisions all share the same namespace! The
	// SchemaLocationCatalog mechanism does not work for 1.0.
	// This has to be removed as soon as the SchemaLocationCatalog mechanism has been refactored.
	private Map<String, String> createSchemaCatalog(SAFEReleaseDescriptor revision) {
		Map<String, String> schemaCatalog = new HashMap<String, String>();
		schemaCatalog.put(revision.getNamespace(), getSchemaLocation(revision));
		return schemaCatalog;
	}

	public String getSchemaLocation(SAFEReleaseDescriptor revision) {
		return SCHEMA_BASE + "_" + getSchemaReleaseNumber(revision) + "." + SCHEMA_EXTENSION; //$NON-NLS-1$//$NON-NLS-2$
	}

	private String getSchemaReleaseNumber(SAFEReleaseDescriptor revision) {
		final String separator = "-"; //$NON-NLS-1$
		SAFEMetaModelVersionData avd = revision.getSAFEVersionData();
		return avd.getMajor() + separator + avd.getMinor() + separator + avd.getRevision();
	}
}